@@ -1,4 +1,5 @@
 import asyncio
+import discord
 import json
 import re
 import shlex
@@ -7,9 +8,8 @@ import urllib.error
 from io import BytesIO
 from red_star.plugin_manager import BasePlugin
 from red_star.rs_errors import CommandSyntaxError, UserPermissionError
-from red_star.rs_utils import respond, is_positive, RSArgumentParser, split_message
+from red_star.rs_utils import respond, is_positive, RSArgumentParser, split_message, prompt_for_confirmation
 from red_star.command_dispatcher import Command
-from discord import InvalidArgument, HTTPException
 from traceback import format_exception, format_exc
 
 
@@ -19,7 +19,7 @@ class BotManagement(BasePlugin):
     author = "medeor413"
     description = "A plugin that allows bot maintainers to interface with core bot options through Discord."
 
-    async def on_dm_message(self, msg):
+    async def on_dm_message(self, msg: discord.Message):
         if msg.author == self.client.user:
             return
         maintainers = [self.client.get_user(x) for x in self.config_manager.config.get("bot_maintainers", [])]
@@ -34,7 +34,7 @@ class BotManagement(BasePlugin):
              doc="Returns the current message latency of the bot.",
              syntax="N/A",
              dm_command=True)
-    async def _ping(self, msg):
+    async def _ping(self, msg: discord.Message):
         await respond(msg, f"**ANALYSIS: Current latency is {round(self.client.latency * 1000)} ms.**")
 
     @Command("Shutdown",
@@ -42,7 +42,7 @@ class BotManagement(BasePlugin):
              category="bot_management",
              dm_command=True,
              bot_maintainers_only=True)
-    async def _shutdown(self, msg):
+    async def _shutdown(self, msg: discord.Message):
         await respond(msg, "**AFFIRMATIVE. SHUTTING DOWN.**")
         raise SystemExit
 
@@ -52,7 +52,7 @@ class BotManagement(BasePlugin):
              category="bot_management",
              bot_maintainers_only=True,
              dm_command=True)
-    async def _update_avatar(self, msg):
+    async def _update_avatar(self, msg: discord.Message):
         try:
             url = msg.content.split(None, 1)[1]
             img = urllib.request.urlopen(url).read()
@@ -68,9 +68,9 @@ class BotManagement(BasePlugin):
         try:
             await self.client.user.edit(avatar=img)
             await respond(msg, "**AVATAR UPDATED.**")
-        except InvalidArgument:
+        except ValueError:
             raise CommandSyntaxError("Image must be a PNG or JPG")
-        except HTTPException:
+        except discord.HTTPException:
             raise UserPermissionError("Cannot change avatar at this time.")
 
     @Command("UpdateName",
@@ -79,7 +79,7 @@ class BotManagement(BasePlugin):
              category="bot_management",
              bot_maintainers_only=True,
              dm_command=True)
-    async def _update_name(self, msg):
+    async def _update_name(self, msg: discord.Message):
         args = msg.clean_content.split()[1:]
         edit_username = False
         if args[0].lower() in ("-p", "--permanent"):
@@ -107,24 +107,27 @@ class BotManagement(BasePlugin):
              category="bot_management",
              bot_maintainers_only=True,
              dm_command=True)
-    async def _activate(self, msg):
-        plgname = msg.content.split()[1]
+    async def _activate(self, msg: discord.Message):
+        plugin_name = msg.content.split()[1]
         try:
             permanent = is_positive(msg.content.split()[2])
         except IndexError:
             permanent = False
-        all_plugins = self.plugin_manager.plugins
-        if plgname in all_plugins:
-            if plgname not in self.plugins:
-                if plgname in self.config_manager.config["disabled_plugins"] and permanent:
-                    self.config_manager.config["disabled_plugins"].remove(plgname)
-                    self.config_manager.save_config()
-                await self.plugin_manager.activate(plgname)
-                await respond(msg, f"**ANALYSIS: Plugin {plgname} was activated successfully.**")
+        all_plugins = self.plugin_manager.plugin_classes
+        if plugin_name in all_plugins:
+            if plugin_name not in self.plugins:
+                if permanent:
+                    disabled_plugins = self.config_manager.get_server_config(
+                            self.guild, "plugin_manager")["disabled_plugins"]
+                    if plugin_name in disabled_plugins:
+                        disabled_plugins.remove(plugin_name)
+                        self.config_manager.save_config()
+                await self.plugin_manager.activate(self.guild, plugin_name)
+                await respond(msg, f"**ANALYSIS: Plugin {plugin_name} was activated successfully.**")
             else:
-                await respond(msg, f"**ANALYSIS: Plugin {plgname} is already activated.**")
+                await respond(msg, f"**ANALYSIS: Plugin {plugin_name} is already activated.**")
         else:
-            await respond(msg, f"**WARNING: Could not find plugin {plgname}.**")
+            await respond(msg, f"**WARNING: Could not find plugin {plugin_name}.**")
 
     @Command("Deactivate",
              doc="Deactivates an active plugin.",
@@ -132,22 +135,25 @@ class BotManagement(BasePlugin):
              category="bot_management",
              bot_maintainers_only=True,
              dm_command=True)
-    async def _deactivate(self, msg):
-        plgname = msg.content.split()[1].lower()
+    async def _deactivate(self, msg: discord.Message):
+        plugin_name = msg.content.split()[1].lower()
         try:
             permanent = is_positive(msg.content.split()[2])
         except IndexError:
             permanent = False
-        if plgname == self.name:
+        if plugin_name == self.name:
             await respond(msg, f"**WARNING: Cannot deactivate {self.name}.**")
-        elif plgname in self.plugins:
-            if plgname not in self.config_manager.config["disabled_plugins"] and permanent:
-                self.config_manager.config["disabled_plugins"].append(plgname)
-                self.config_manager.save_config()
-            await self.plugin_manager.deactivate(plgname)
-            await respond(msg, f"**ANALYSIS: Plugin {plgname} was deactivated successfully.**")
+        elif plugin_name in self.plugins:
+            if permanent:
+                disabled_plugins = self.config_manager.get_server_config(
+                        self.guild, "plugin_manager")["disabled_plugins"]
+                if plugin_name not in disabled_plugins:
+                    disabled_plugins.append(plugin_name)
+                    self.config_manager.save_config()
+            await self.plugin_manager.deactivate(self.guild, plugin_name)
+            await respond(msg, f"**ANALYSIS: Plugin {plugin_name} was deactivated successfully.**")
         else:
-            await respond(msg, f"**ANALYSIS: Plugin {plgname} is not active.**")
+            await respond(msg, f"**ANALYSIS: Plugin {plugin_name} is not active.**")
 
     @Command("ReloadPlugin",
              doc="Reloads a plugin module, refreshing code changes.",
@@ -155,7 +161,7 @@ class BotManagement(BasePlugin):
              category="bot_management",
              bot_maintainers_only=True,
              dm_command=True)
-    async def _reload_plugin(self, msg):
+    async def _reload_plugin(self, msg: discord.Message):
         plgname = msg.content.split()[1].lower()
         if plgname == self.name:
             await respond(msg, f"**WARNING: Cannot deactivate {self.name}.**")
@@ -169,54 +175,61 @@ class BotManagement(BasePlugin):
              category="bot_management",
              bot_maintainers_only=True,
              dm_command=True)
-    async def _list_plugins(self, msg):
-        active_plgs = ", ".join(self.plugins.keys())
-        if not active_plgs:
-            active_plgs = "None."
-        all_plgs = list(self.plugin_manager.plugins)
-        inactive_plgs = ", ".join([x for x in all_plgs if x not in self.plugins])
-        if not inactive_plgs:
-            inactive_plgs = "None."
-        await respond(msg, f"**ANALYSIS: Plugins are as follows:**```\nActive: {active_plgs}\n"
-                           f"Inactive: {inactive_plgs}\n```")
+    async def _list_plugins(self, msg: discord.Message):
+        active_plugins = ", ".join(self.plugins.keys())
+        if not active_plugins:
+            active_plugins = "None."
+        all_plugins = list(self.plugin_manager.plugin_classes)
+        inactive_plugins = ", ".join([x for x in all_plugins if x not in self.plugins])
+        if not inactive_plugins:
+            inactive_plugins = "None."
+        await respond(msg, f"**ANALYSIS: Plugins are as follows:**```\nActive: {active_plugins}\n"
+                           f"Inactive: {inactive_plugins}\n```")
 
     @Command("GetConfig",
              doc="Gets the config value at the specified path. Use <server> to fill in the server ID.\n"
-                 "Use --file to view a different configuration file, such as music_player.json.",
-             syntax="(path/to/value) [-f/--file file]",
-             category="bot_management",
-             bot_maintainers_only=True,
-             dm_command=True)
-    async def _get_config(self, msg):
-        args = msg.clean_content.split()[1:]
-        if args[0] in ("-f", "--file"):
-            args.pop(0)
-            filename = args.pop(0)
-            try:
-                conf_dict = self.config_manager.plugin_config_files[filename]
-            except KeyError:
-                raise CommandSyntaxError(f"Config file {filename} does not exist.")
-        else:
-            conf_dict = self.config_manager.config.copy()
+                 "Use --plugin-storage to view a plugin's storage instead of the config.\n"
+                 "Bot maintainers only: Use --default to view default configurations, or --global to view the "
+                 "non-server configuration.",
+             syntax="(path/to/value) [-s/--plugin-storage plugin_name] [-d/--default] [-g/--global]",
+             category="bot_management", perms="administrator")
+    async def _get_config(self, msg: discord.Message):
+        parser = RSArgumentParser()
+        parser.add_argument("command")
+        parser.add_argument("path", default="")
+        mutex_group = parser.add_mutually_exclusive_group()
+        mutex_group.add_argument("-s", "--plugin-storage")
+        mutex_group.add_argument("-d", "--default-config", action="store_true")
+        mutex_group.add_argument("-g", "--global-config", action="store_true")
 
-        try:
-            path = args[0]
-        except IndexError:
-            path = ""
-        try:
-            path = path.replace("<server>", str(msg.guild.id))
-        except AttributeError:
-            path = path.replace("<server>", "default")
-        if path.startswith("/"):
-            path = path[1:]
-        path_list = path.split("/")
-        conf_dict.pop("token", None)  # Don't wanna leak that by accident!
+        args = parser.parse_args(shlex.split(msg.content))
+
+        if args.plugin_storage:
+            try:
+                conf_dict = self.plugins[args.plugin_storage].storage.copy()
+            except KeyError:
+                raise CommandSyntaxError(f"Plugin with name {args.plugin_storage} does not exist.")
+        elif args.default_config:
+            if not self.config_manager.is_maintainer(msg.author):
+                raise UserPermissionError
+            conf_dict = self.config_manager.config["default"].copy()
+        elif args.global_config:
+            if not self.config_manager.is_maintainer(msg.author):
+                raise UserPermissionError
+            conf_dict = self.config_manager.config["global"].copy()
+            conf_dict.pop("token", None)  # Don't want to leak that by accident!
+        else:
+            conf_dict = self.config_manager.config[str(msg.guild.id)].copy()
+
+        if args.path.startswith("/"):
+            args.path = args.path[1:]
+        path_list = args.path.split("/")
 
         for k in path_list:
             if not k:
                 break
             try:
-                conf_dict = self._list_or_dict_subscript(conf_dict, k)
+                conf_dict = self._list_or_dict_index(conf_dict, k)
             except KeyError:
                 raise CommandSyntaxError(f"Key {k} does not exist.")
             except IndexError:
@@ -227,7 +240,7 @@ class BotManagement(BasePlugin):
                 raise CommandSyntaxError(f"{args.path} is not a valid path!")
 
         res = json.dumps(conf_dict, indent=2, sort_keys=True)
-        for split_msg in split_message(f"**ANALYSIS: Contents of {path}:**```JSON{res}```"):
+        for split_msg in split_message(f"**ANALYSIS: Contents of /{args.path}:**```JSON\n{res}\n```"):
             await respond(msg, split_msg)
 
     @Command("SetConfig",
@@ -242,19 +255,23 @@ class BotManagement(BasePlugin):
              category="bot_management",
              bot_maintainers_only=True,
              dm_command=True)
-    async def _set_config(self, msg):
+    async def _set_config(self, msg: discord.Message):
         parser = RSArgumentParser()
+        parser.add_argument("command")
         parser.add_argument("path")
         parser.add_argument("value", nargs="*")
-        parser.add_argument("-f", "--file", type=str)
         parser.add_argument("-t", "--type", choices=("null", "bool", "int", "float", "str", "list", "dict", "json"),
                             type=str.lower)
-        group = parser.add_mutually_exclusive_group()
-        group.add_argument("-r", "--remove", action="store_true")
-        group.add_argument("-a", "--append", action="store_true")
-        group.add_argument("-k", "--addkey")
+        operation_group = parser.add_mutually_exclusive_group()
+        operation_group.add_argument("-r", "--remove", action="store_true")
+        operation_group.add_argument("-a", "--append", action="store_true")
+        operation_group.add_argument("-k", "--addkey")
+        source_group = parser.add_mutually_exclusive_group()
+        source_group.add_argument("-s", "--plugin-storage")
+        source_group.add_argument("-d", "--default-config", action="store_true")
+        source_group.add_argument("-g", "--global-config", action="store_true")
 
-        args = parser.parse_args(shlex.split(msg.clean_content)[1:])
+        args = parser.parse_args(shlex.split(msg.content))
 
         type_converters = {
             "null": lambda x: None,
@@ -268,20 +285,24 @@ class BotManagement(BasePlugin):
             "json": json.loads
         }
 
-        if args.file:
+        if args.plugin_storage:
             try:
-                conf_dict = self.config_manager.plugin_config_files[args.file]
+                conf_dict = self.plugins[args.plugin_storage].storage
             except KeyError:
-                raise CommandSyntaxError(f"Config file {args.file} does not exist.")
+                raise CommandSyntaxError(f"Plugin with name {args.plugin_storage} does not exist.")
+        elif args.default_config:
+            if not self.config_manager.is_maintainer(msg.author):
+                raise UserPermissionError
+            conf_dict = self.config_manager.config["default"]
+        elif args.global_config:
+            if not self.config_manager.is_maintainer(msg.author):
+                raise UserPermissionError
+            conf_dict = self.config_manager.config["global"]
         else:
-            conf_dict = self.config_manager.config.copy()
+            conf_dict = self.config_manager.config[str(msg.guild.id)]
 
         try:
             path = args.path
-            try:
-                path = path.replace("<server>", str(msg.guild.id))
-            except AttributeError:
-                path = path.replace("<server>", "default")
             if path.startswith("/"):
                 path = path[1:]
             path_list = path.split("/")
@@ -298,7 +319,7 @@ class BotManagement(BasePlugin):
             raise CommandSyntaxError("Missing path to config value.")
         for k in path_list:
             try:
-                conf_dict = self._list_or_dict_subscript(conf_dict, k)
+                conf_dict = self._list_or_dict_index(conf_dict, k)
             except KeyError:
                 raise CommandSyntaxError(f"Key {k} does not exist.")
             except IndexError:
@@ -307,7 +328,7 @@ class BotManagement(BasePlugin):
                 raise CommandSyntaxError(f"{args.path} is not a valid path!")
 
         try:
-            orig = self._list_or_dict_subscript(conf_dict, final_key)
+            orig = self._list_or_dict_index(conf_dict, final_key)
         except KeyError:
             raise CommandSyntaxError(f"Key {final_key} does not exist.")
         except IndexError:
@@ -324,19 +345,19 @@ class BotManagement(BasePlugin):
                 conf_dict.pop(int(final_key))
             else:
                 raise CommandSyntaxError("Path does not lead to a collection!")
-            await respond(msg, f"**ANALYSIS: Config value {args.path} deleted successfully.**")
+            await respond(msg, f"**ANALYSIS: Config value /{path} deleted successfully.**")
 
         elif args.append:
             if isinstance(conf_dict[final_key], list):
                 conf_dict[final_key].append(value)
-                await respond(msg, f"**ANALYSIS: {value} appended to {path} successfully.**")
+                await respond(msg, f"**ANALYSIS: {value} appended to /{path} successfully.**")
             else:
                 raise CommandSyntaxError("Path does not lead to a list!")
 
         elif args.addkey:
             if isinstance(conf_dict[final_key], dict):
                 conf_dict[final_key][args.addkey] = value
-                await respond(msg, f"**ANALYSIS: Key {args.addkey} with value {value} added to {path} "
+                await respond(msg, f"**ANALYSIS: Key {args.addkey} with value {value} added to /{path} "
                                    f"successfully.**")
             else:
                 raise CommandSyntaxError("Path does not lead to a dict!")
@@ -360,13 +381,13 @@ class BotManagement(BasePlugin):
              category="debug",
              perms={"manage_guild"},
              dm_command=True)
-    async def _last_error(self, msg):
+    async def _last_error(self, msg: discord.Message):
         try:
             args = msg.clean_content.split(None, 1)[1].lower()
         except IndexError:
             raise CommandSyntaxError("No error context specified.")
         if args == "command":
-            e = self.client.command_dispatcher.last_error
+            e = self.command_dispatcher.last_error
         elif args == "event":
             e = self.plugin_manager.last_error
         elif args == "unhandled":
@@ -374,11 +395,85 @@ class BotManagement(BasePlugin):
         else:
             raise CommandSyntaxError("Invalid error context.")
         if e:
-            excstr = "\n".join(format_exception(*e))
-            await respond(msg, f"**ANALYSIS: Last error in context {args}:** ```Python\n{excstr}\n```")
+            traceback_str = "\n".join(format_exception(*e))
+            await respond(msg, f"**ANALYSIS: Last error in context {args}:** ```Python\n{traceback_str}\n```")
         else:
             await respond(msg, f"**ANALYSIS: No error in context {args}.**")
 
+    @Command("PurgeServerConfig",
+             doc="Removes the configuration data for a server that the bot is no longer on.",
+             syntax="(server ID, or all)",
+             category="bot_management",
+             bot_maintainers_only=True)
+    async def _purge_server_config(self, msg: discord.Message):
+        try:
+            target_guild_id = msg.clean_content.split(None, 1)[1]
+            if not target_guild_id.isdigit():
+                if target_guild_id.lower() == "all":
+                    target_guild_id = None
+                else:
+                    raise CommandSyntaxError("Invalid server ID.")
+        except KeyError:
+            raise CommandSyntaxError("Missing server ID.")
+
+        if target_guild_id:
+            guild_storage_dir = self.config_manager.storage_path / target_guild_id
+            guild_in_config = target_guild_id in self.config_manager.config
+            guild_in_storage = guild_storage_dir.exists()
+
+            if not guild_in_config and not guild_in_storage:
+                await respond(msg, f"**WARNING: Guild with ID {target_guild_id} not found in config or storage."
+                                   f"Confirm the ID is correct.**")
+                return
+
+            confirmed = await prompt_for_confirmation(msg, f"Really delete ALL data for server {target_guild_id}?")
+            if not confirmed:
+                await msg.delete()
+                return
+
+            if guild_on := self.client.get_guild(int(target_guild_id)):
+                prompt_text = f"The bot is currently active on server {target_guild_id} ({guild_on.name}). Purging " \
+                              f"the configuration of an active server is EXTREMELY DANGEROUS. Continue anyways?"
+                confirmed = await prompt_for_confirmation(msg, prompt_text)
+                if not confirmed:
+                    await msg.delete()
+                    return
+
+            if guild_in_config:
+                self.config_manager.config.pop(target_guild_id)
+                self.config_manager.save_config()
+            if guild_in_storage:
+                for file in guild_storage_dir.iterdir():
+                    file.unlink()
+                guild_storage_dir.rmdir()
+                self.config_manager.storage_files[target_guild_id] = {}
+
+            await respond(msg, f"**ANALYSIS: The configuration for server ID {target_guild_id} has been purged.**")
+        else:  # Purge all unused data.
+            confirmed = await prompt_for_confirmation(msg, f"Really delete ALL data for servers the bot is not a "
+                                                           f"member of?")
+            if not confirmed:
+                await msg.delete()
+                return
+
+            guilds_to_ignore = {str(x.id) for x in self.client.guilds}
+            guilds_to_ignore.add("global")
+            guilds_to_ignore.add("default")
+
+            for server_id in tuple(self.config_manager.config.keys()):
+                if server_id not in guilds_to_ignore:
+                    self.config_manager.config.pop(server_id)
+            self.config_manager.save_config()
+            for server_folder in self.config_manager.storage_path.iterdir():
+                if server_folder.name not in guilds_to_ignore:
+                    for file in server_folder.iterdir():
+                        file.unlink()
+                    server_folder.rmdir()
+                    self.config_manager.storage_files[server_folder.name] = {}
+
+            await respond(msg, "**ANALYSIS: The configuration data for all servers the bot is not a member of has "
+                               "been purged.**")
+
     @Command("Execute", "Exec", "Eval",
              doc="Executes the given Python code. Be careful, you can really break things with this!\n"
                  "Provided variables are `ct` (shorthand for asyncio.create_task) and `self.res` "
@@ -388,7 +483,7 @@ class BotManagement(BasePlugin):
              bot_maintainers_only=True,
              run_anywhere=True,
              dm_command=True)
-    async def _execute(self, msg):
+    async def _execute(self, msg: discord.Message):
         try:
             arg = re.split(r"\s+", msg.content, 1)[1]
             t_match = re.match(r"`([^\n\r`]+)`|```.*?\s+(.*)```", arg, re.DOTALL)
@@ -411,13 +506,13 @@ class BotManagement(BasePlugin):
             self.res = self.EmptySentinel()
 
     @staticmethod
-    def _list_or_dict_subscript(obj, key):
+    def _list_or_dict_index(obj: dict | list, key: str):
         """
-        Helper function to use key as a key if obj is a dict, or as an int index if list.
+        Helper function to use key as a key if obj is a dict, or as an int index if a list.
 
-        :param obj: The object to subscript.
-        :param key: The key to subscript with.
-        :return: The subscripted object.
+        :param obj: The object to retrieve from.
+        :param key: The index to retrieve.
+        :return: The object at the index.
         """
         if isinstance(obj, dict):
             return obj[key]
@@ -428,5 +523,5 @@ class BotManagement(BasePlugin):
 
     class EmptySentinel:
         """
-        Just a simple class to function as a sentinel value for Execute's result variable.
+        An empty sentinel value for the result variable in Execute.
         """
